package org.hmn.jrsd.rt;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.nio.charset.Charset;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import simpleorm.dataset.SFieldMeta;
import simpleorm.dataset.SRecordInstance;
import simpleorm.dataset.SRecordMeta;
import simpleorm.sessionjdbc.SSessionJdbc;

/**
 * Utilities to use with Simpleorm. For best BLOB/CLOB support, the database 
 * must support PreparedStatement.setBytes() resp. PreparedStatement.setString()
 * which are used to initialize empty BLOB resp. CLOB columns. 
 * @author Karadimas
 * Cree par Harry Karadimas a 3 sept. 08
 */
public class SormUtils
{
  static Logger lg = Logger.getLogger(SormUtils.class);
  static final String NL = System.getProperty("line.separator");
  
  /**
   * Return true if the field's meta designates one of the primary keys of the 
   * record.
   * @param fm
   * @return
   */
  public static boolean isPartOfPrimaryKey(SFieldMeta fm)
  {
    SRecordMeta<?> rm = fm.getRecordMeta();
    SFieldMeta[] pks = rm.getPrimaryKeys();
    for (SFieldMeta pk:pks) {
      if (pk.equals(fm)) return true;
    }
    //not found, return false
    return false;
  }

  /**
   * Convenienc method to execute a ddl statement. The ThreadLocalSession is
   * used. Nothing in begin()ned or committed().
   * @return The number that was returned by executeUpdate() 
   * @param sql
   * @throws SQLException
   */
  static public int ddl(String sql)
  throws SQLException
  {
    SSessionJdbc session = SSessionJdbc.getThreadLocalSession();
    Connection cxn = session.getJdbcConnection();
    Statement st = cxn.createStatement();
    try {
      lg.debug("Executing DDL SQL \n'"+sql+"'");
      int n = st.executeUpdate(sql);
      //st gets closed in the "finally" block
      return n;      
    }
    finally {
      st.close();
    }
  }

  /**
   * Return true if the table exists (in any schema, any catalog of the 
   * database).
   * @param tableName
   * @return
   * @throws SQLException
   */
  static public boolean tableExists(String tableName)
  throws SQLException
  {
    return tableExists(null, null, tableName);
  }
  
  /**
   * Return true if the table exists in the given schema and the given 
   * catalog. Uses the current ThreadLocalSession.
   * @param catalog
   * @param schema
   * @param tableName
   * @return
   * @throws SQLException
   */
  static public boolean tableExists(
      String catalog, String schema, String tableName)
  throws SQLException
  {
    SSessionJdbc session = SSessionJdbc.getThreadLocalSession();
    Connection cxn = session.getJdbcConnection();
    DatabaseMetaData md = cxn.getMetaData();    
    ResultSet rs = md.getTables(catalog, schema, tableName, null);
    int i = 0;
    while (rs.next()) i++;
    return i == 1;
  }

  /**
   * Return true if the column exists in the given table, independently of
   * any schema and catalog. Use this method on a small database, when there
   * are not different schemas and different catalogs.
   * @param tableName
   * @param columnName
   * @return
   * @throws SQLException
   */
  static public boolean columnExists(String tableName, String columnName)
  throws SQLException
  {
    return columnExists(null, null, tableName, columnName);
  }
  
  /**
   * Return true if the column exists in the given table, in the given schema
   * and the given catalog.
   * @param catalog
   * @param schema
   * @param tableName
   * @param columnName
   * @return
   * @throws SQLException
   */
  static public boolean columnExists(
      String catalog, String schema, String tableName, String columnName)
  throws SQLException
  {
    SSessionJdbc session = SSessionJdbc.getThreadLocalSession();
    Connection cxn = session.getJdbcConnection();
    DatabaseMetaData md = cxn.getMetaData();    
    ResultSet rs = md.getColumns(catalog, schema, tableName, columnName);
    int i = 0;
    while (rs.next()) i++;
    return i == 1;
  }

  /**
   * Query column size, ignoring catalog and schema. Useful for small databases
   * where there is only one schema / one catalog.
   * @param cxn
   * @param tableName
   * @param columnName
   * @return
   * @throws SQLException
   */
  public static int getColumnSize(
      Connection cxn, String tableName, String columnName)
  throws SQLException
  {
    return getColumnSize(cxn, null, null, tableName, columnName);
  }
  
  /**
   * Query column size
   * @param cxn
   * @param catalog
   * @param schema
   * @param tableName
   * @param columnName
   * @return The column's size, or -1 if the column was not found.
   * @throws SQLException
   */
  public static int getColumnSize(
      Connection cxn, 
      String catalog, String schema, String tableName, String columnName)
  throws SQLException
  {
    int csz = -1;
    DatabaseMetaData md = cxn.getMetaData();
    ResultSet rs = md.getColumns(catalog, schema, tableName, columnName);
    while (rs.next()) {
      csz = rs.getInt("COLUMN_SIZE");
    }
    return csz;
  }
  
  /**
   * Load the given resource as a String, using the platform's default charset.
   * This is fine if your resource is purely ASCII; however explicit use
   * of "UTF-8" is more portable.
   * @see #loadResourceAsString(String, String)
   * @param resourceName
   * @return
   */
  static public String loadResourceAsString(String resourceName)
  {
    try {
      return loadResourceAsString(resourceName, Charset.defaultCharset().name());
    }
    catch (UnsupportedEncodingException e) {
      throw new RuntimeException("Unexpected exception", e);
    }
  }
  
  /**
   * Read the given resource as a String. Useful to load SQL files that are
   * present as resources. The given encoding should be "UTF-8" for better
   * portability.
   * @param resourceName
   * @return
   */
  static public String loadResourceAsString(String resourceName, String encoding)
  throws UnsupportedEncodingException
  {
    InputStream ins = SormUtils.class.getResourceAsStream(resourceName);
    if (ins == null) throw new RuntimeException("Resource not found  : " + resourceName);
    InputStreamReader insr = new InputStreamReader(ins, encoding);
    StringWriter sw = new StringWriter();
    int c = 0;
    try {
      while ((c = insr.read()) >= 0) sw.write(c);
      insr.close();
    }
    catch (IOException ex) {
      throw new RuntimeException(ex);
    }
    return sw.toString();
  }
  
  /**
   * From a string that contains several sql ddl statements, create a list
   * of strings. The convention used here is that each sql statement is 
   * separated from the others by a line that contains a single semicolon (";").
   * @param sqls
   * @return A list of DDL statements
   */
  static public List<String> separateSqlInstructions(String sqls)
  {
    StringBuffer sb = new StringBuffer();
    ArrayList<String> strings = new ArrayList<String>();
    BufferedReader brdr = new BufferedReader(new StringReader(sqls));
    String line;
    try {
    while ((line = brdr.readLine()) != null) {
      if (line.trim().equals(";")) {
        strings.add(sb.toString());
        sb = new StringBuffer();
      }
      else {
        if (sb.length() > 0) sb.append('\n'); 
        sb.append(line); 
      }
    }//while
    }
    catch (IOException ex) {
      throw new RuntimeException(ex);
    }
    String lastSql = sb.toString();
    if (!isBlank(lastSql)) strings.add(lastSql);
    return strings;
  }

  /**
   * Execute the sql ddl statements that are contained in the resource. The 
   * convention used here is that each sql statement is separated from the 
   * others by a line that contains a single semicolon (";").
   * @param cxn
   * @param resourceSql
   * @param encoding
   * @param lgr The logger where traces should be emitted, in the form of
   *   "Executing DDS SQL '...'". Ignored if null.
   * @param lvl
   * @throws IOException
   * @throws SQLException
   */
  public static void executeResourceSqls(
      Connection cxn, String resourceSql, String encoding, Logger lgr, Level lvl)
  throws IOException, SQLException
  {
    if (lgr != null) lgr.log(lvl, "Loading sql from resource '" + resourceSql + "'");
    String sqls = loadResourceAsString(resourceSql, encoding);
    List<String> sqlStrings = separateSqlInstructions(sqls);
    for (String sql : sqlStrings) {
      Statement st = cxn.createStatement();
      if (lgr != null) lgr.log(lvl, "Executing DDL SQL \n'"+sql+"'");
      st.execute(sql);
      st.close();
    }
  }
  
  /**
   * @param str
   * @return True if str contains only spaces, tabs or newline chars (\r , \n).
   */
  static public boolean isBlank(String str)
  {
    int len = str.length();
    for (int i = 0; i < len; i++) {
      char c = str.charAt(i);
      if (c != ' ' & c != '\t' & c != '\r' & c != '\n') return false;
    }
    return true;
  }
  
  /**
   * Query something that returns an Integer. Parameters are possible, but
   * try to avoid null parameters, as these usually only work if you use
   * "IS NULL" in where clauses, not when using "=?". For more elaborate
   * queries, use an SqlExpr.
   * @param cxn
   * @param sql
   * @param params
   * @return
   * @throws SQLException
   */
  public static Integer queryInteger(
      Connection cxn, String sql, Object... params)
  throws SQLException
  {
    PreparedStatement ps = null;
    try {
      ps = cxn.prepareStatement(sql);
      for (int i = 0; i < params.length; i++) {
        Object param = params[i];
        if (param == null) ps.setNull(i+1, Types.VARCHAR);
        else ps.setObject(i+1, param);
      }
      ResultSet rs = ps.executeQuery();      
      Integer ir = null;
      if (rs.next()) {
        int i = rs.getInt(1);
        if (rs.wasNull()) ir = null;
        else ir = new Integer(i);        
      }//while
      rs.close();
      return ir;      
    }
    finally {
      if (ps != null) ps.close();
    }
  }  

  /**
   * Do an extra update, for a field that is not in the SRecord, typically
   * a BLOB or CLOB. Example usage :
   * <pre>
   * SormUtils.updateExtra(s, rec, "MY_BLOB = empty_blob()"); //Oracle blob initialization
   * </pre>
   * @param s
   * @param rec
   * @param setClause
   * @throws SQLException
   */
  public static void updateExtra(
      SSessionJdbc s, SRecordInstance rec, String setClause)
  throws SQLException
  {
    RecordIdentifier recId = RecordIdentifier.makeRecordIdentifier(rec);
    AndExpr expr = new AndExpr();    
    expr.addAllEqClauses(recId.getPkNames(), recId.getPkValues());
    String sql = "UPDATE \"" + recId.getTableName() + "\" SET " + setClause + 
                 " WHERE " + expr.getText();
    PreparedStatement ps = s.getJdbcConnection().prepareStatement(sql);
    try {
      expr.setParameters(1, ps);
      ps.executeUpdate();
    }
    finally {
      ps.close();
    }
  }
  
  /**
   * Updates the column by forcing the use of PreparedStatement.setString() on
   * the column. SimpleOrm relies on PreparedStatement.setObject() to set the
   * value. Using PreparedStatement.setString() is useful for example to store
   * values in CLOB fields easily.
   * @param s
   * @param rec
   * @param columnName
   * @param value
   * @throws SQLException
   */
  public static void updateColumnAsString(
      SSessionJdbc s, SRecordInstance rec, String columnName, String value)
  throws SQLException
  {
    RecordIdentifier recId = RecordIdentifier.makeRecordIdentifier(rec);
    AndExpr expr = new AndExpr();    
    expr.addAllEqClauses(recId.getPkNames(), recId.getPkValues());
    String sql = "UPDATE \"" + recId.getTableName() + "\" SET \"" + 
                 columnName + "\"=?" + " WHERE " + expr.getText();
    PreparedStatement ps = s.getJdbcConnection().prepareStatement(sql);
    try {
      if (value == null) ps.setNull(1, Types.VARCHAR);
      else ps.setString(1, value);
      expr.setParameters(2, ps);
      ps.executeUpdate();
    }
    finally {
      ps.close();
    }
  }
  
  /**
   * Updates the column by forcing the use of PreparedStatement.setString() on
   * the column. SimpleOrm relies on PreparedStatement.setObject() to set the
   * value. Using PreparedStatement.setBytes() is useful for example to store
   * values in BLOB fields easily.
   * @param s
   * @param rec
   * @param columnName
   * @param value
   * @throws SQLException
   */
  public static void updateColumnAsBytes(
      SSessionJdbc s, SRecordInstance rec, String columnName, byte[] value)
  throws SQLException
  {
    RecordIdentifier recId = RecordIdentifier.makeRecordIdentifier(rec);
    AndExpr expr = new AndExpr();    
    expr.addAllEqClauses(recId.getPkNames(), recId.getPkValues());
    String sql = "UPDATE \"" + recId.getTableName() + "\" SET \"" + 
                 columnName + "\"=?" + " WHERE " + expr.getText();
    PreparedStatement ps = s.getJdbcConnection().prepareStatement(sql);
    try {
      if (value == null) ps.setNull(1, Types.BLOB);
      else ps.setBytes(1, value);
      expr.setParameters(2, ps); // set the rest of the parameters starting at 2
      ps.executeUpdate();
    }
    finally {
      ps.close();
    }
  }
  
  /**
   * Calls {@link #selectExtra(SSessionJdbc, SRecordInstance, String, ResultSetColVisitable, boolean)}
   * with forUpdate = true
   * 
   * @see #selectExtra(SSessionJdbc, SRecordInstance, String, ResultSetColVisitable, boolean)
   * 
   * @param s
   * @param rec
   * @param selClause
   * @param visitor
   * @throws SQLException 
   * @throws SQLException
   * @throws ResultSetColVisitException 
   */
  public static void selectExtra(
      SSessionJdbc s, SRecordInstance rec, String selClause,
      ResultSetColVisitable visitor)
  throws SQLException, ResultSetColVisitException
  {
    selectExtra(s, rec, selClause, visitor, true);
  }
  
  /**
   * Select extra things (columns that are not in the meta) for a record 
   * (typically a BLOB or CLOB), using its primary keys, and organize the visit 
   * of each record for the given visitor.
   * 
   * @param s
   * @param rec
   * @param selClause
   * @param visitor
   * @param forUpdateLock If true, " FOR UPDATE" is appended at the end of the
   *   query, which has the effect of locking the selected row(s)
   * @throws SQLException 
   * @throws SQLException
   * @throws ResultSetColVisitException 
   */
  public static void selectExtra(
      SSessionJdbc s, SRecordInstance rec, String selClause,
      ResultSetColVisitable visitor, boolean forUpdateLock)
  throws SQLException, ResultSetColVisitException
  {
    RecordIdentifier recId = RecordIdentifier.makeRecordIdentifier(rec);
    AndExpr expr = new AndExpr();    
    expr.addAllEqClauses(recId.getPkNames(), recId.getPkValues());
    String sql = "SELECT " + selClause + " FROM \"" + recId.getTableName() + 
    "\" WHERE " + expr.getText();
    if (forUpdateLock) sql = sql + " FOR UPDATE";
    PreparedStatement ps = s.getJdbcConnection().prepareStatement(sql);
    try {
      expr.setParameters(1, ps);
      ResultSet rs = ps.executeQuery();
      while (rs.next()) visitor.visit(rs);
      rs.close();      
    }
    finally {
      ps.close();
    }
  }
  
  /**
   * Read the bytes from the Blob. Uses a BufferedInputStream for better
   * performance.
   * @param blb
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static byte[] readFromBlob(Blob blb)
  throws SQLException, IOException
  {
    InputStream ins = blb.getBinaryStream();
    BufferedInputStream bis = new BufferedInputStream(ins);
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int b;
    while ((b = bis.read()) >= 0) bos.write(b);
    bis.close();
    ins.close();    
    return bos.toByteArray();
  }
  
  /**
   * Write the bytes to the Blob. Uses a buffered output stream for better
   * performance.
   * @param bytes
   * @param blb
   * @throws SQLException
   * @throws IOException
   */
  public static void writeToBlob(byte[] bytes, Blob blb)
  throws SQLException, IOException
  {
    OutputStream os = blb.setBinaryStream(0);
    BufferedOutputStream bos = new BufferedOutputStream(os);
    bos.write(bytes);
    bos.close();
    os.close();
  }
  
  /**
   * Read the chars from the Clob. Uses a BufferedReader for better 
   * performance.
   * @param clb
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static String readFromClob(Clob clb)
  throws SQLException, IOException
  {
    StringBuffer sb = new StringBuffer();
    Reader rdr = clb.getCharacterStream();
    BufferedReader brdr = new BufferedReader(rdr);
    int c;
    while ((c = brdr.read()) >= 0) sb.append((char)c);
    brdr.close();
    rdr.close();
    return sb.toString();
  }
  
  /**
   * Write the String's chars to the Clob. Uses a BufferedWriter for better
   * performance.
   * @param str
   * @param clb
   * @throws SQLException
   * @throws IOException
   */
  public static void writeToClob(String str, Clob clb)
  throws SQLException, IOException
  {
    char[] chars = str.toCharArray();
    Writer wr = clb.setCharacterStream(0);
    BufferedWriter bwr = new BufferedWriter(wr);
    bwr.write(chars);
    bwr.close();
    wr.close();
  }

  /**
   * Calls {@link #readFromBlob(SSessionJdbc, SRecordInstance, String, boolean)}
   * with lck = true
   * 
   * @param s
   * @param rec
   * @param blobColumnName
   * @return
   * @throws ResultSetColVisitException
   * @throws SQLException
   * @throws IOException
   */
  public static byte[] readFromBlob(
      SSessionJdbc s, SRecordInstance rec, String blobColumnName)
  throws ResultSetColVisitException, SQLException, IOException
  {
    return readFromBlob(s, rec, blobColumnName, true);
  }
  
  /**
   * Read the bytes from the Blob, in a column that is in the same table as
   * rec. The primary key(s) of the records is(are) used to select the Blob. 
   * On most databases the record must already have been read with a
   * FOR UPDATE lock.
   * @param s The session
   * @param rec The record whose primary keys must be used.
   * @param blobColumnName The table's column that contains the Blob 
   * @param lck If true, the row is re-locked using FOR UPDATE clause.
   * @return
   * @throws ResultSetColVisitException
   * @throws SQLException
   * @throws IOException
   */
  public static byte[] readFromBlob(
      SSessionJdbc s, SRecordInstance rec, String blobColumnName, boolean lck)
  throws ResultSetColVisitException, SQLException, IOException
  {
    BlobCollectVisitor visitor = new BlobCollectVisitor();
    SormUtils.selectExtra(s, rec, blobColumnName, visitor, lck);
    Blob clb = visitor.getBlob(); 
    if (clb == null) return null;
    byte[] bytes = SormUtils.readFromBlob(clb);
    return bytes;
  }
    
  /**
   * Calls {@link #writeToBlob(SSessionJdbc, SRecordInstance, String, byte[], boolean)}
   * with lck = true.
   * 
   * @param s
   * @param rec
   * @param blobColumnName
   * @param bytes
   * @throws IOException
   * @throws SQLException
   * @throws ResultSetColVisitException
   */
  public static void writeToBlob(
      SSessionJdbc s, SRecordInstance rec, String blobColumnName, byte[] bytes)
  throws IOException, SQLException, ResultSetColVisitException
  {
    writeToBlob(s, rec, blobColumnName, bytes, true);
  }
  
  /**
   * Write to blob. If the blog was not initialized, updateColumnAsBytes() is
   * called.
   * @see SormUtils#updateExtra(SSessionJdbc, SRecordInstance, String)
   * @param s
   * @param rec
   * @param blobColumnName
   * @param bytes
   * @throws ResultSetColVisitException 
   * @throws SQLException 
   * @throws Exception
   */
  public static void writeToBlob(
      SSessionJdbc s, SRecordInstance rec, String blobColumnName, 
      byte[] bytes, boolean lck)
  throws IOException, SQLException, ResultSetColVisitException
  {
    BlobCollectVisitor visitor = new BlobCollectVisitor();
    SormUtils.selectExtra(s, rec, blobColumnName, visitor, lck);
    Blob blb = visitor.getBlob(); 
    if (blb == null) updateColumnAsBytes(s, rec, blobColumnName, bytes);
    else SormUtils.writeToBlob(bytes, blb);
  }

  /**
   * Calls {@link #readFromClob(SSessionJdbc, SRecordInstance, String, boolean)}
   * with lck = true.
   * 
   * @param s
   * @param rec
   * @param clobColumnName
   * @return
   * @throws IOException
   * @throws SQLException
   * @throws ResultSetColVisitException
   */
  public static String readFromClob(
      SSessionJdbc s, SRecordInstance rec, String clobColumnName)
  throws IOException, SQLException, ResultSetColVisitException
  {
    return readFromClob(s, rec, clobColumnName, true);
  }
  
  /**
   * Read from Blob, using a visitor.
   * @param s
   * @param rec
   * @param clobColumnName
   * @param lck If true, re-locks the row using FOR UPDATE clause
   * @return
   * @throws IOException
   * @throws SQLException
   * @throws ResultSetColVisitException
   */
  public static String readFromClob(
      SSessionJdbc s, SRecordInstance rec, String clobColumnName, boolean lck)
  throws IOException, SQLException, ResultSetColVisitException
  {
    ClobCollectVisitor visitor = new ClobCollectVisitor();
    SormUtils.selectExtra(s, rec, clobColumnName, visitor);
    Clob clb = visitor.getClob(); 
    if (clb == null) return null;
    String str = SormUtils.readFromClob(clb);
    return str;    
  }

  public static void writeToClob(
      SSessionJdbc s, SRecordInstance rec, String clobColumnName, String str)
  throws IOException, SQLException, ResultSetColVisitException
  {
    writeToClob(s, rec, clobColumnName, str, true);
  }
  
  /**
   * Write to clob. If the Blog was not initialized, updateColumnAsString() will
   * be called.
   * @see SormUtils#updateExtra(SSessionJdbc, SRecordInstance, String)
   * 
   * @param s
   * @param rec
   * @param clobColumnName
   * @param str
   * @throws ResultSetColVisitException 
   * @throws SQLException 
   * @throws Exception
   */
  public static void writeToClob(
      SSessionJdbc s, SRecordInstance rec, String clobColumnName, 
      String str, boolean lck)
  throws IOException, SQLException, ResultSetColVisitException
  {
    ClobCollectVisitor visitor = new ClobCollectVisitor();
    SormUtils.selectExtra(s, rec, clobColumnName, visitor, lck);
    Clob clb = visitor.getClob();     
    if (clb == null) updateColumnAsString(s, rec, clobColumnName, str); 
    else SormUtils.writeToClob(str, clb);    
  }
  
  /**
   * Dump SQL exception and its chained exceptions to the given StringBuffer.
   * The system's line.separator is used for newlines.
   * @param ex
   * @param lgr
   */
  public static void dumpSqlExceptions(SQLException ex, StringBuffer sb)
  {
    SQLException nex = ex;
    while (nex != null) {
      if (sb.length() > 0) sb.append(NL);
      sb.append("SQLException " + ex.getMessage() + 
                ", error code : " + ex.getErrorCode() + 
                ", sql state : " + ex.getSQLState());    
      nex = ex.getNextException();      
    }
  }

  /**
   * Collect all the chained SQL Exceptions and return them in a string. 
   * @param ex
   * @return
   */
  public static String getChainedDumps(SQLException ex)
  {
    StringBuffer sb = new StringBuffer();
    dumpSqlExceptions(ex, sb);
    return sb.toString();
  }
  
}
