package org.hmn.jrsd.rt.dbops;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.sql.DataSource;

import org.hmn.jrsd.rt.dbops.DbListBlobOperation;
import org.hmn.jrsd.rt.dbops.DbListOperation;
import org.hmn.jrsd.rt.dbops.DbObjectBlobOperation;
import org.hmn.jrsd.rt.dbops.DbObjectOperation;
import org.hmn.jrsd.rt.dbops.DbOperation;
import org.hmn.jrsd.rt.dbops.DbUpdateBlobOperation;
import org.hmn.jrsd.rt.dbops.DbUpdateOperation;

import simpleorm.sessionjdbc.SSessionJdbc;

/*
 * 
 * Cree par Harry Karadimas a 2 avr. 08
 */
/**
 * This class has utilities to execute one or more operations, inside a 
 * transaction, with a session
 * created for this operation. Most methods are static in this object
 */
public class DbOps
{
  private ArrayList<DbOperation> opList = new ArrayList<DbOperation>();
  
  public DbOps() {
  }
  
  public DbOps(Collection<DbOperation> c) { opList.addAll(c); }
  
  public List<DbOperation> getOperations() { return opList; }
  
  public void addOperation(DbOperation op) { opList.add(op); }
  
  /** execute a select operation that involves a blob, inside a transaction */
  public static Object executeObjectOp(
      DbObjectBlobOperation op, DataSource ds, String connectionName)
  throws SQLException, IOException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);
    try {
      session.begin();      
      Object r = op.execute();
      session.commit();
      return r;
    }
    finally {
      session.detachFromThread();
      session.close();
    }
  }

  /**
   * Execute the operation that involves a blob and report the timing
   * @param op
   * @param ds
   * @param connectionName
   * @param tr
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static Object executeObjectOp(
      DbObjectBlobOperation op, DataSource ds, String connectionName,
      TimingReporter tr)
  throws SQLException, IOException
  {
    long tic = System.currentTimeMillis();
    Object r = executeObjectOp(op, ds, connectionName);
    long tac = System.currentTimeMillis();
    if (tr != null) tr.reportTiming(connectionName, tac - tic);
    return r;        
  }
  
  /** execute an operation */
  public static Object executeObjectOp(
      DbObjectOperation op, DataSource ds, String connectionName)
  throws SQLException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);
    try {
      session.begin();
      Object r = op.execute();
      session.commit();
      return r;
    }
    finally {
      session.detachFromThread();
      session.close();
    }
  }
  
  /**
   * Execute an operation and report its timing
   * @param op
   * @param ds
   * @param connectionName
   * @param tr
   * @return
   * @throws SQLException
   */
  public static Object executeObjectOp(
      DbObjectOperation op, DataSource ds, String connectionName,
      TimingReporter tr)
  throws SQLException
  {
    long tic = System.currentTimeMillis();
    Object r = executeObjectOp(op, ds, connectionName);
    long tac = System.currentTimeMillis();
    if (tr != null) tr.reportTiming(connectionName, tac - tic);
    return r;    
    
  }
  
  /** execute a series of operations 
   * @throws IOException */
  public static Object[] executeOps(
      DbOps op, DataSource ds, String connectionName)
  throws SQLException, IOException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);    
    try {
      session.begin();
      Object[] r = op.executeOperations();
      session.commit();
      return r;
    }
    finally {
      session.close();
    }
  }
  
  /**
   * Execute all operations in sequence, return results as array :
   * Result of operation x is at position x in the array.
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public Object[] executeOperations()
  throws SQLException, IOException
  {
    Object[] results = new Object[opList.size()];
    for (int i = 0; i < opList.size(); i++) {
      DbOperation op = opList.get(i);
      switch (op.getOpType()) {
      case DbOperation.LIST_OP:
        List<?> lst = ((DbListBlobOperation)op).execute();
        results[i] = lst;
        break;
      case DbOperation.OBJECT_OP:
        Object obj = ((DbObjectBlobOperation)op).execute();
        results[i] = obj;
        break;
      case DbOperation.UPDATE_OP:
        Integer res = ((DbUpdateBlobOperation)op).execute();
        results[i] = res;
        break;
      }
    }//for
    return results;
  }

  /** execute an update operation */
  public static int execute(
      DbUpdateOperation op, DataSource ds, String connectionName)
  throws SQLException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);
    try {
      session.begin();
      int n = op.execute();
      session.commit();
      return n;
    }
    finally {
      session.close();
    }
  }

  public static int execute(
      DbUpdateOperation op, DataSource ds, String connectionName,
      TimingReporter tr)
  throws SQLException
  {
    long tic = System.currentTimeMillis();
    int r = execute(op, ds, connectionName);
    long tac = System.currentTimeMillis();
    if (tr != null) tr.reportTiming(connectionName, tac - tic);
    return r;    
  }

  /** execute an update operation that involves a Blob */
  public static int execute(DbUpdateBlobOperation op, DataSource ds, String connectionName)
  throws SQLException, IOException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);
    try {
      session.begin();
      int n = op.execute();
      session.commit();
      return n;
    }
    finally {
      session.close();
    }
  }
  
  /**
   * Execute an update operation that involves a blob, and report the timing
   * of the operation.
   * @param op
   * @param ds
   * @param connectionName
   * @param tr
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static int execute(
      DbUpdateBlobOperation op, DataSource ds, String connectionName,
      TimingReporter tr)
  throws SQLException, IOException
  {
    long tic = System.currentTimeMillis();
    int r = execute(op, ds, connectionName);
    long tac = System.currentTimeMillis();
    if (tr != null) tr.reportTiming(connectionName, tac - tic);
    return r;    
  }

  /** execute a selection operation */
  public static List<?> executeListOp(DbListOperation op, DataSource ds, String connectionName)
  throws SQLException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);    
    try {
      session.begin();
      List<?> r = op.execute();
      session.commit();
      return r;
    }
    finally {
      session.close();
    }
  }

  /** execute a selection operation and report its timing */
  public static List<?> executeListOp(
      DbListOperation op, DataSource ds, String connectionName, 
      TimingReporter tr)
  throws SQLException, IOException
  {
    long tic = System.currentTimeMillis();
    List<?> r = executeListOp(op, ds, connectionName);
    long tac = System.currentTimeMillis();
    if (tr != null) tr.reportTiming(connectionName, tac - tic);
    return r;
  }
  
  /** execute a selection operation that involves a blob */
  public static List<?> executeListOp(
      DbListBlobOperation op, DataSource ds, String connectionName)
  throws SQLException, IOException
  {
    SSessionJdbc session = SSessionJdbc.open(ds, connectionName);
    try {
      session.begin();
      List<?> r = op.execute();
      session.commit();
      return r;
    }
    finally {
      session.close();
    }
  }

  /** 
   * execute a selection operation that involves a blob, and report its timing 
   */
  public static List<?> executeListOp(
      DbListBlobOperation op, DataSource ds, String connectionName, 
      TimingReporter tr)
  throws SQLException, IOException
  {
    long tic = System.currentTimeMillis();
    List<?> r = executeListOp(op, ds, connectionName);
    long tac = System.currentTimeMillis();
    if (tr != null) tr.reportTiming(connectionName, tac - tic);
    return r;
  }
  
}
