package org.hmn.jrsd.rt.oracle;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Set;

import oracle.sql.BLOB;
import oracle.sql.CLOB;

import org.hmn.jrsd.rt.AndExpr;
import org.hmn.jrsd.rt.BinOpExpr;
import org.hmn.jrsd.rt.QColumn;
import org.hmn.jrsd.rt.SDriverOracleNoLock;
import org.hmn.jrsd.rt.SormUtils;

import simpleorm.dataset.SFieldMeta;
import simpleorm.dataset.SRecordInstance;
import simpleorm.dataset.SRecordMeta;
import simpleorm.sessionjdbc.SSessionJdbc;

/*
 * 
 * Cree par Harry Karadimas a 15 dc. 2006
 */
/**
 * Oracle-specific LOB utilities, for LOB management. These are set apart,
 * because LOB handling is not very jdbc specific in Oracle. These utilies
 * help to write correctly to LOBs. Beware that the row must have been
 * locked using "FOR UPDATE" clause, before retrieving the LOB value. Also the
 * LOB must exist, I.E. the value of the LOB column must be the result of the 
 * built-in function "EMPTY_BLOB()" or "EMPTY_CLOB()", not NULL. Some methods
 * use deprecated values because it enables the use of old oracle jdbc drivers
 * (v 8). 
 * @author karadimas
 */
public class OracleLOBUtils
{
  public static final int DEF_BLOCK_SIZE = 4096;

  //---
  //--- BLOB Operations
  //---
  
  /**
   * Write all the InputStream to the BLOB, piece by piece. The InputStream
   * is closed when this method returns.
   * @param in
   * @param b
   * @param blockSize
   * @throws SQLException
   * @throws IOException
   */
  public static void writeToBLOB(InputStream in, BLOB b, int blockSize)
  throws SQLException, IOException
  {
    if (b == null) throw new NullPointerException("BLOB argument is null");    
    OutputStream out = b.getBinaryOutputStream();    
    int nread;
    byte[] block = new byte[blockSize];
    while ((nread = in.read(block)) >= 0) out.write(block, 0, nread);
    out.close();    
    in.close();
  }
  
  /**
   * Write to a BLOB, one chunk at a time, to avoid driver saturation.
   * @param data
   * @param b
   * @param blockSize
   * @throws SQLException
   * @throws IOException
   */
  public static void writeToBLOB(byte[] data, BLOB b, int blockSize)
  throws SQLException, IOException
  {
    ByteArrayInputStream in = new ByteArrayInputStream(data);
    writeToBLOB(in, b, blockSize);    
  }
  
  /**
   * Write to a BLOB using the default block size.
   * @param data
   * @param b
   * @throws SQLException
   * @throws IOException
   * @see #DEF_BLOCK_SIZE
   */
  public static void writeToBLOB(byte[] data, BLOB b)
  throws SQLException, IOException
  {
    writeToBLOB(data, b, DEF_BLOCK_SIZE);
  }
  
  /**
   * Read byte data from the BLOB and write it to the output stream, block by 
   * block. The output stream is closed when this method returns.
   * @param b
   * @param out
   * @param blockSize
   * @throws SQLException
   * @throws IOException
   */
  public static void readFromBLOB(BLOB b, OutputStream out, int blockSize)
  throws SQLException, IOException
  {
    if (b == null) throw new NullPointerException("BLOB argument is null");
    if (out == null) throw new NullPointerException("OutputStream argument is null");
    InputStream in = b.getBinaryStream();
    byte[] block = new byte[blockSize];
    int nread;
    while ((nread = in.read(block)) >= 0) out.write(block, 0, nread);
    in.close();
    out.close();
  }
  
  /**
   * Read from BLOB, one chunk at a time, to avoid saturation.
   * @param b
   * @param blockSize
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static byte[] readFromBLOB(BLOB b, int blockSize)
  throws SQLException, IOException
  {
    ByteArrayOutputStream out = new ByteArrayOutputStream(blockSize);
    readFromBLOB(b, out, blockSize);
    return out.toByteArray();
  }
  
  /**
   * Read from BLOB, using the default size.
   * @param b
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static byte[] readFromBLOB(BLOB b)
  throws SQLException, IOException
  {
    return readFromBLOB(b, DEF_BLOCK_SIZE);
  }
  
  //---
  //--- CLOB specific operations
  //---
  
  /**
   * Copy character data from the reader to the CLOB, one piece at a time,
   * to avoid saturation. The Reader is closed by this method.
   * @param br
   * @param c
   * @param blockSize
   */
  public static void writeToCLOB(Reader in, CLOB c, int blockSize)
  throws SQLException, IOException
  {
    Writer out = c.getCharacterOutputStream();
    char[] buf = new char[blockSize];
    int nread;
    while ((nread = in.read(buf)) >= 0) out.write(buf, 0, nread);
    out.close();
    in.close();
  }
  
  /**
   * Write the string to the CLOB.
   * @param data
   * @param b
   * @throws SQLException
   * @throws IOException
   */
  public static void writeToCLOB(String data, CLOB c, int blockSize)
  throws SQLException, IOException
  {
    writeToCLOB(new StringReader(data), c, blockSize);
  }
  
  public static void writeToCLOB(String data, CLOB c)
  throws SQLException, IOException
  {
    writeToCLOB(data, c, DEF_BLOCK_SIZE);
  }
  
  /**
   * Read data from CLOB and write it to the Writer. The Writer is closed
   * when this method returns.
   * @param c
   * @param out
   * @param blockSize
   * @throws SQLException
   * @throws IOException
   */
  public static void readFromCLOB(CLOB c, Writer out, int blockSize)
  throws SQLException, IOException
  {
    char[] block = new char[blockSize];
    Reader in = c.getCharacterStream();
    int nread;
    while ((nread = in.read(block)) >= 0) out.write(block, 0, nread);
    in.close();
    out.close();
  }
  
  /**
   * Read a string from a CLOB, one block at a time.
   * @param c
   * @param blockSize
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static String readFromCLOB(CLOB c, int blockSize)
  throws SQLException, IOException
  {
    StringWriter out = new StringWriter(blockSize);
    readFromCLOB(c, out, blockSize);
    return out.toString();
  }
  
  /**
   * Read a string from a CLOB, one block at a time.
   * @param c
   * @return
   * @throws SQLException
   * @throws IOException
   */
  public static String readFromCLOB(CLOB c)
  throws SQLException, IOException
  {
    return readFromCLOB(c, DEF_BLOCK_SIZE);
  }
  
  /**
   * Lock row explicitly, for SimpleORM, useful when locking has been disabled
   * through the {@link SDriverOracleNoLock} custom driver.
   * Be careful with that one, the row will remain locked until the end of the 
   * transaction. Deadlock <i>do</i> occur in Oracle.
   * Beware that SimpleOrm does not allow NULL values for primary keys (just
   * like Oracle and others...).
   * @param rec
   * @throws IOException
   */
  public static void sLockRow(SRecordInstance rec)
  throws IOException
  {
    String tableName = (String) rec.getMeta().getTableName();
    String sqlPrefix = "SELECT * FROM "+QColumn.quoteIfNecessary(tableName);
    AndExpr clauses = new AndExpr();
    SRecordMeta<?> meta = rec.getMeta();
    Set<String> fieldNames = meta.getFieldNames();
    for (Iterator<String> iterator = fieldNames.iterator(); iterator.hasNext();) {
      String fieldName = iterator.next();
      SFieldMeta fieldMeta = meta.getField(fieldName);
      if (SormUtils.isPartOfPrimaryKey(fieldMeta)) {
        Object val = rec.getObject(fieldMeta);
        BinOpExpr expr = new BinOpExpr(fieldName, "=", val);
        clauses.add(expr);
      }
    }//
    String sql = sqlPrefix + " WHERE " + clauses.getText() + " FOR UPDATE";
    SSessionJdbc session = SSessionJdbc.getThreadLocalSession();
    session.rawUpdateDB(sql, clauses.getParameters().toArray());    
  }
    
  /**
   * Lock row explicitly. Useful for tables that can't be used in in simple 
   * orm, because their composite primary key can contain a null value
   * (nulls are not supported for primary key values in simple orm).
   * @param table
   * @param pkNames
   * @param pkVals
   * @throws IOException
   */
  public static void lockRow(String table, String[] pkNames, Object[] pkVals)
  throws IOException
  {
    StringBuffer sb = new StringBuffer();
    sb.append("SELECT * FROM ");
    sb.append(QColumn.quoteIfNecessary(table));
    AndExpr clauses = new AndExpr();
    for (int i = 0; i < pkNames.length; i++) {
      String pkName = pkNames[i];
      Object pkVal = pkVals[i];
      BinOpExpr expr = new BinOpExpr(pkName, "=", pkVal);
      clauses.add(expr);
    }//for
    sb.append(clauses.getText());
    sb.append(" FOR UPDATE");
    SSessionJdbc session = SSessionJdbc.getThreadLocalSession();
    session.rawUpdateDB(sb.toString(), clauses.getParameters().toArray());    
  }
  
}
