package org.hmn.jrsd.rt.rows;

import java.io.Serializable;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/*
 * 
 * Cree par Harry Karadimas a 5 fvr. 2007
 */
/**
 * Results from a Jdbc select query, to be serialized for client VM use
 * (typical use : server VM makes the Jdbc queries, client VM is the web server)
 * The JdbcResultRows, and other client-side objects are all constructed from
 * the this object that contains the "raw" results. You can "decorate" objects
 * on the client side with formatting attributes, that can be reused by your
 * templating code (Stringtemplate or "st" is quite practical for this).
 * <br>
 * On the server side :
 * <pre>
 *   JdbcResults ress = new JdbcResult();
 *   ress.init(rs);
 * </pre>
 * On the client side :
 * <pre>
 *   JdbcResults ress = ... //get JdbcResults from a server call
 *   ress.clientInit(); //do the client-side typical initialization
 * </pre>
 * @author karadimas
 */
public class JdbcResults
implements Serializable
{
  private static final long serialVersionUID = -5947901139372623011L;
  
  JdbcResultColumns columns;
  ArrayList<Object[]> rawResults;
  private JdbcResultRows rows;
  private String encoding;
  private HashMap<String,CellFormat> formatsByLabel = 
    new HashMap<String, CellFormat>();
  private HashMap<Serializable,Serializable> attributes = 
    new HashMap<Serializable, Serializable>();
  
  /**
   * Declare a cell format for a given column. This method is typically called 
   * on the client side. If you declare formats on the server side (which is
   * not recommended), your formats must be serializable. 
   * @param columnLabel
   * @param fmt
   */
  public void putFormat(String columnLabel, CellFormat fmt)
  {
    formatsByLabel.put(columnLabel, fmt);
  }
  
  public CellFormat getFormat(String columnLabel)
  {
    return (CellFormat) formatsByLabel.get(columnLabel);
  }
  
  /**
   * Constructor with an existing ResultSet. 
   */
  public JdbcResults(ResultSet rs)
    throws SQLException
  {
    columns = new JdbcResultColumns(rs);
    rawResults = new ArrayList<Object[]>();
    int cc = columns.getColumnCount();
    while (rs.next()) {
      Object[] resRow = new Object[cc];
      for (int i = 1; i <= cc; i++) resRow[i-1] = rs.getObject(i);
      rawResults.add(resRow);
    }
    rs.close();
  }
    
  public JdbcResultColumns getColumns() {
    return columns;
  }
   
  public Object[] getRowObjects(int index) {
    return (Object[]) rawResults.get(index);
  }

  public List<Object[]> getRawResults() { return rawResults; }
  
  /**
   * This method is called once the object has been transferred on the client.   
   */
  public void clientInit()
  {
    rows = new JdbcResultRows(this);    
  }
  
  /**
   * Client-side initialization code, that specifies a specialized cell class
   * to use for result cells.    
   */
  public void clientInit(Class<? extends JdbcResultCell> jdbcResultCellClass)
  {
    rows = new JdbcResultRows(this, jdbcResultCellClass);    
  }
  
  public int getRowCount()
  {
    return rawResults.size();
  }
  
  public JdbcResultRows getRows() {
    return rows;
  }
  
  public String getEncoding() {
    return encoding;
  }
  
  public void setEncoding(String encoding) {
    this.encoding = encoding;
  }
  
  /**
   * Get the map of extra attributes you want to pass with this results object
   * (like duration of the execution, memory used, ... whatever).
   * You can pass extra information using this map, however all your keys
   * and values must be Serializable.
   * @return
   */
  public HashMap<Serializable, Serializable> getAttributes() {
    return attributes;
  }
}
